
#-------------------------------------------------------------------------------
# OFORTH : assembler macros
#-------------------------------------------------------------------------------

#ifndef __USER_LABEL_PREFIX__
#define __USER_LABEL_PREFIX__ _
#endif

#define CONCAT1(a, b)  a##b
#define CONCAT2(a, b)  CONCAT1(a,b)
#define SYM(x)         CONCAT2(__USER_LABEL_PREFIX__, x)

#define __CHAR      '
#define char(x)     $CONCAT2(__CHAR, x)

#ifdef __x8632__
#include "x8632.h"
#endif

#ifdef __x8664__
#include "x8664.h"
#endif

#ifdef __arm32__
#include "arm32.h"
#endif

#define NATURAL_LOW      CONCAT2(LIT, 0x3FFFFFFF)
#define NATURAL_SHIFT    CONCAT2(LIT, 30)
#define MSKCLASSID       CONCAT2(LIT, 0x00000FFF)
#define MSKSTATE         CONCAT2(LIT, 0xFFFFF000)
#define MSKSIZEALL       CONCAT2(LIT, 0x0003F000)
#define MSKSIZEHIDDEN    CONCAT2(LIT, 0x003C0000)
#define MSKNOTATTGC      CONCAT2(LIT, 0x00400000)
#define MSKATTGC         CONCAT2(LIT, 0xFFBFFFFF)
#define MSKMODEINIT      CONCAT2(LIT, 0x00800000)
#define MSKMODERUN       CONCAT2(LIT, 0xFF7FFFFF)
#define MSKMUTABLEOBJ    CONCAT2(LIT, 0x01000000)
#define MSKNOTMUTABLE    CONCAT2(LIT, 0xFEFFFFFF)
#define MSKMUTABLEORINIT CONCAT2(LIT, 0x01800000)
#define MSKALLOCATED     CONCAT2(LIT, 0x02000000)
#define MSKBUFFER        CONCAT2(LIT, 0x04000000)
#define MSKPERSISTENT    CONCAT2(LIT, 0x08000000)
#define MSKPERSISTALLOC  CONCAT2(LIT, 0x0A000000)
#define MSKARRAYGC       CONCAT2(LIT, 0x10000000)
#define MSKFINALIZE      CONCAT2(LIT, 0x20000000)
#define MSKIMMEDIATE     CONCAT2(LIT, 0x40000000)
#define MSKNOTFINALIZE   CONCAT2(LIT, 0xDFFFFFFF)
#define MSKFREE          CONCAT2(LIT, 0x80000000)
#define MSKNOTFREE       CONCAT2(LIT, 0x7FFFFFFF)
#define MSKPERSISTFREE   CONCAT2(LIT, 0x88000000)

#define C1               CONCAT2(LIT, CELLSIZE)
#define C2               CONCAT2(LIT, CELLSIZE * 2)
#define C3               CONCAT2(LIT, CELLSIZE * 3)
#define C4               CONCAT2(LIT, CELLSIZE * 4)
#define C5               CONCAT2(LIT, CELLSIZE * 5)
#define C6               CONCAT2(LIT, CELLSIZE * 6)
#define C7               CONCAT2(LIT, CELLSIZE * 7)
#define C8               CONCAT2(LIT, CELLSIZE * 8)
#define CFRAMEUP         $2

#define nil             CONCAT2(LIT, 0x0)

#define ifIntegerJmp(x,l)       test(x, $1); jne l
#define ifNIntegerJmp(x,l)      test(x, $1); je  l

#define mulc(x)         sal(CELLSHIFT, x)
#define divc(x)         sar(CELLSHIFT, x)
#define check(x)        test(x, x)
#define setzero(x)      xor(x, x)
#define setnil(x)       xor(x, x)
#define isNull(x)       check(x)

#define to_cint(x)      dec(x); sar($1, x)
#define to_oint(x)      sal($1, x); inc(x)

#define ZERO            $1
#define ONE             $3
#define TWO             $5
#define THREE           $7
#define false           ZERO
#define true            ONE

#define top             at(ds)
#define second          at1(ds)
#define third           at2(ds)
#define fourth          at3(ds)
#define drop            add(C1, ds)
#define drop2           add(C2, ds)
#define drop3           add(C3, ds)
#define drop4           add(C4, ds)
#define popds(x)        ldm(x, top); drop
#define pushds(x)       stm(x, bt1(ds));    sub(C1, ds)
#define pushdsimm(x)    movimm(x, bt1(ds)); sub(C1, ds)
#define push2ds(x,y)    sub(C2, ds); stm(x, at1(ds)); stm(y, top)

#define self            at(rs)
#define pushrs(x)       push(x)
#define poprs(x)        pop(x)
#define decrs(x)        add(x, rs)

// Memory pages and GC
#define MSKSIZE         $12
#define MSKHSIZE        $18

#define FIRST_PAGE      $0
#define NEXT_PAGE       $2

#define GCFREE          $0x00
#define GCFINALIZE      $0x01
#define GCNOTUSED       $0x02
#define GCNEW           $0x04
#define GCNOTUSEDNEW    $0x06
#define GCUSED          $0x08
#define GCUSEDIMMUT     $0x10
#define GCNOTREMOVE     $0x20
#define GCNOTREMOVEUSED $0x28

#define opnbfree(x)     at(x)
#define opnextstate(x)  at1(x)
#define opnext(x)       at2(x)
#define opstates(x)     at3(x)
#define opobjects(x)    at4(x)

// Tasks
#define krs(x)          at(x)
#define kinner(x)       at1(x)
#define kstate(x)       at2(x)
#define kip(x)          at3(x)
#define knatural(x)     at6(x)
#define krsbase(x)      at7(x)
#define kvars(x)        at11(x)
#define kdssize(x)      at12(x)
#define krunnable(x)    at13(x)
#define kfloat(x)       at14(x)

// CThreads
#define THREAD_RUNNING  CONCAT2(LIT, 0x00000000)
#define RESTRICT_GC     CONCAT2(LIT, 0x00100000)
#define GLOBAL_GC       CONCAT2(LIT, 0x01000000)
#define ASK_GC          $0x1

#define GC_NOGC         $0
#define GC_SWEEPING     $2

#define thds            at(thread)
#define thstate         at1(thread)
#define thdsbase        at2(thread)
#define thcurpages      at3(thread)
#define thallpages      at4(thread)
#define thraws          at5(thread)
#define thrsoforth      at7(thread)
#define thrsbase        at8(thread)
#define thrspause       at9(thread)
#define thippause       at10(thread)
#define thtask          at11(thread)
#define thdsup          at13(thread)
#define thvm            at14(thread)
#define thgcphase       at15(thread)
#define thgcticks       at16(thread)
#define thgcmarkstack   at17(thread)
#define thgcmarkbase    at18(thread)
#define thgcmarkup      at19(thread)
#define thgcfreeindex   at20(thread)

#define savetds         stm(ds, thds)
#define restds          ldm(ds, thds)

#define IDNil           CONCAT2(LIT, 2)
#define IDInteger       CONCAT2(LIT, 3)
#define IDClass         CONCAT2(LIT, 5) 
#define IDFloat         CONCAT2(LIT, 6)

// Raw
#define rnext(x)        at(x)
#define robj(x)         at1(x)
#define rdata(x)        at2(x)

// Objects
#define objptr(x)       at(x)
 
// Pointers
#define pvalue(x)       at1(x)

// Words
#define wname(x)        at2(x)

// Class
#define clinitfunc(x)   at4(x)
#define clfinalfunc(x)  at5(x)
#define classPtr(x)     at6(x)
#define clsize(x)       at7(x)
#define clsuper(x)      at9(x)

// Blocks
#define CLOSURESIZE     $3

#define blcode(x)       at1(x)
#define blnbvars(x)     at2(x)
#define blvars(x)       at3(x)

// Functions
#define fncode(x)       at3(x) 

// NodeMethods
#define nodenext(x)     at1(x)
#define nodecode(x)     at2(x)
#define nodeimpl(x)     at3(x)

// Methods
#define virtualTable(x) at3(x)
#define classImpl(x)    at5(x)

// Implements
#define implcode(x)     at2(x)

// Dynamic library
#define libtype(x)      at4(x)

// Dynamic procedures
#define dynhandle(x)    at3(x)
#define dynnbparam(x)   at4(x)
#define dynlib(x)       at5(x)
#define dynreturn(x)    at7(x)

// Constants
#define cvalue(x)       at3(x)

// Variables
#define varindex(x)     at3(x)

// Naturals
#define nbuf(x)         at1(x)
#define nfree(x)        at2(x)
#define nsign(x)        at3(x)

// Symbols
#define symbuf(x)       at2(x)

// Strings
#define sbuf(x)		    at1(x)
#define sfree(x)        at2(x)
#define salloc(x)       at3(x)

// Lists
#define lbuf(x)         at1(x)
#define lfree(x)        at2(x)
#define lalloc(x)       at3(x)

// Files
#define FILE_UTF8       TWO
#define fhandle(x)      at1(x)
#define fname(x)        at2(x)
#define fencoding(x)    at3(x)

// TryNodes
#define tlevel(x)       at1(x)
#define taddr(x)        at2(x)

// Resources
#define rclosed(x)      at5(x)

//*****************************************************************************
#if defined(__OWIN32__)
#define OINTTOF(x)      sub($1, x)

#define ABI_PRE(n) 
#define ABI_RET

#define func(x)     ALIGNED;.globl SYM(x);.def SYM(x);.scl 2;.type 32;.endef;SYM(x):;
#endif

//*****************************************************************************
#if defined(__OLNX32__)
#define OINTTOF(x)      sub($1, x)

#define ABI_PRE(n) 
#define ABI_RET

#if defined(__arm__)
#define func(x)                  .globl SYM(x); .type SYM(x), %function; SYM(x):;
#else
#define func(x)         ALIGNED; .globl SYM(x); .type SYM(x), @function; SYM(x):;
#endif

#endif

//*****************************************************************************
#if defined(__OMAC32__)
#define OINTTOF(x)      sub($1, x)

#define ABI_PRE(n)       mov(rs, r1); sub(n, rs); sub(ABI_ALIGN, rs); and(ABI_MASK, rs); add(n, rs); sub(rs, r1); stm(r1, at(rs)); 
#define ABI_RET          add((rs), rs); 

#define func(x)         .align 8; .globl SYM(x); SYM(x):;
#endif

//*****************************************************************************
LDeb:    .ascii     "%x \0";
LPoint:  .ascii     ".\0";
LDebugd: .ascii     "Debug reached : %d %x \12\0";
LDebugs: .ascii     "Debug reached : %s \12\0";
#define  deb(x)     pusha; call2(printf, CONCAT2(LIT, LDeb), x); popa
#define  dp         pusha; call1(printf, CONCAT2(LIT, LPoint)); popa
#define  debugd(x)  pusha; call3(printf, CONCAT2(LIT, LDebugd), x, x); popa
#define  debugs(x)  pusha; call2(printf, CONCAT2(LIT, LDebugs), x); popa


